﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Data.SQLite;
using System.Text;
using ThinkGeo.MapSuite.Core;

namespace WorldMapKitDataExtractor
{
    public class Extractor
    {
        private string inputDataPath;
        private string outputDataPath;
        private string inputDataPrj;
        private string boundaryPrj;
        private bool preserveCountryLevelData;

        public Extractor(string inputDataPath, string outputDataPath, string inputDataPrj = null, string boundaryPrj = null, bool preserveCountryLevelData = true)
        {
            this.inputDataPath = inputDataPath;
            this.outputDataPath = outputDataPath;
            this.inputDataPrj = inputDataPrj;
            this.boundaryPrj = boundaryPrj;
            this.preserveCountryLevelData = preserveCountryLevelData;
        }

        public Action<string> UpdateStatus { get; set; }

        public string InputDataPath
        {
            get { return inputDataPath; }
            set { inputDataPath = value; }
        }

        public string OutputDataPath
        {
            get { return outputDataPath; }
            set { outputDataPath = value; }
        }

        public string InputDataPrj
        {
            get { return inputDataPrj; }
            set { inputDataPrj = value; }
        }

        public string BoundaryPrj
        {
            get { return boundaryPrj; }
            set { boundaryPrj = value; }
        }

        public bool PreserveCountryLevelData
        {
            get { return preserveCountryLevelData; }
            set { preserveCountryLevelData = value; }
        }

        public void ExtractDataByShape(IEnumerable<Feature> boundaryFeatures)
        {
            UpdateStatus("Beginning to extract world map kit data by shape file.");

            var projectedBoundaryFeatures = ProjectedBoundaryFeatures(boundaryFeatures);
            SQLiteConnection.CreateFile(outputDataPath);
            SQLiteConnection targetDatabaseConnection = new SQLiteConnection($"Data Source={outputDataPath};Version=3;");
            targetDatabaseConnection.Open();
            SQLiteConnection sourceDatabaseConnection = new SQLiteConnection($"Data Source={inputDataPath};Version=3;");
            sourceDatabaseConnection.Open();

            CopyDataToABuffer(ExtentHelper.GetBoundingBoxOfItems(projectedBoundaryFeatures), sourceDatabaseConnection);
            ProcessBaselandByShapes(projectedBoundaryFeatures, targetDatabaseConnection);

            UpdateStatus("Executing vacuum for output database.");
            ExecuteNonQueryCommand("VACUUM;", targetDatabaseConnection);

            targetDatabaseConnection.Close();
            targetDatabaseConnection.Dispose();
            UpdateStatus("Done extracting world map kit data!");
        }

        private void CopyDataToABuffer(RectangleShape boundingBox, SQLiteConnection sourceDatabaseConnection)
        {
            var views = GetViews(sourceDatabaseConnection);

            //OSM WorldMapKit contains several views and view indices that we want to grab as well for our OSMLayer, so we create temporary tables of the data from these within the boundingbox

            foreach (View view in views)
            {
                UpdateStatus($"Creating temporary table for {view.Name}");

                StringBuilder createNewTempTableFromViewCommand = new StringBuilder();

                createNewTempTableFromViewCommand.AppendFormat("CREATE TEMP TABLE {0} AS ", view.Name);
                createNewTempTableFromViewCommand.AppendFormat("SELECT * FROM {0} WHERE id in (", view.Name);
                createNewTempTableFromViewCommand.AppendFormat("SELECT id FROM {0} WHERE id in (", view.BaseTable);
                createNewTempTableFromViewCommand.AppendFormat("SELECT id from idx_{0}_geometry WHERE maxx < {1} AND maxy < {2} AND minx > {3} AND miny > {4}));", view.Name, boundingBox.UpperRightPoint.X, boundingBox.UpperRightPoint.Y, boundingBox.LowerLeftPoint.X, boundingBox.LowerLeftPoint.Y);

                ExecuteNonQueryCommand(createNewTempTableFromViewCommand.ToString(), sourceDatabaseConnection);
            }

            Proj4Projection proj4 = new Proj4Projection(inputDataPrj, inputDataPrj);
            GeographyUnit geographyUnit = proj4.GetInternalGeographyUnit();

            //get the features from every table within the bounding box and add them to the temporary output file
            var tables = GetTables(sourceDatabaseConnection);
            foreach (var table in tables)
            {
                UpdateStatus($"Copying {table.TableName} table to output database.");
                SqliteFeatureSource source = new SqliteFeatureSource($@"Data Source={inputDataPath};Version=3;", table.TableName);
                source.Open();

                Collection<Feature> intersectFeatures = GetFeaturesIntersects(source, boundingBox, ReturningColumnsType.AllColumns);

                if (preserveCountryLevelData && (table.TableName.Equals("osm_baseland_polygon") || table.TableName.Equals("osm_baseland1m_polygon")))
                    intersectFeatures = source.GetAllFeatures(ReturningColumnsType.AllColumns);

                Collection<string> targetTables = SqliteFeatureSource.GetTableNames($@"Data Source={outputDataPath};Version=3;");

                if (!targetTables.Contains(table.TableName))
                    SqliteFeatureSource.CreateTable($@"Data Source={outputDataPath};Version=3;", table.TableName, table.Columns, geographyUnit);

                SqliteFeatureSource tempSource = new SqliteFeatureSource($@"Data Source={outputDataPath};Version=3;", table.TableName);
                tempSource.Open();
                tempSource.BeginTransaction();

                int counter = 0;
                foreach (var feature in intersectFeatures)
                {
                    UpdateStatus($"Adding new feature: {feature.Id}.");
                    tempSource.AddFeature(feature);
                    counter++;
                }
                UpdateStatus("Commiting sqlite database transaction.");
                tempSource.CommitTransaction();
                tempSource.Close();
            }
        }

        private IEnumerable<Feature> ProjectedBoundaryFeatures(IEnumerable<Feature> boundaryFeatures)
        {
            Proj4Projection proj = new Proj4Projection(boundaryPrj, inputDataPrj);
            proj.Open();
            foreach (Feature feature in boundaryFeatures)
            {
                yield return proj.ConvertToExternalProjection(feature);
            }
            proj.Close();
        }

        private void ProcessBaselandByShapes(IEnumerable<Feature> boundaryFeatures, SQLiteConnection connection)
        {
            UpdateStatus("Processing baseland by shapes.");
            if (!preserveCountryLevelData) return;

            MultipolygonShape boundaryShape = new MultipolygonShape();
            foreach (Feature feature in boundaryFeatures)
            {
                boundaryShape.Polygons.Add(feature.GetBoundingBox().ToPolygon());
            }

            Feature boundaryFeature = new Feature(boundaryShape);
            boundaryFeature = SqlTypesGeometryHelper.MakeValid(boundaryFeature);


            SqliteFeatureSource baselandFeatureSource = new SqliteFeatureSource(connection.ConnectionString, "osm_baseland_polygon", "id", "geometry");
            SqliteFeatureSource baseland1mFeatureSource = new SqliteFeatureSource(connection.ConnectionString, "osm_baseland1m_polygon", "id", "geometry");

            baselandFeatureSource.Open();
            baseland1mFeatureSource.Open();

            UpdateStatus("Querying baseland features from osm_baseland_polygon table.");
            Collection<Feature> baselandFeatures = baselandFeatureSource.SpatialQuery(boundaryShape, QueryType.Intersects, ReturningColumnsType.NoColumns);

            UpdateStatus("Querying baseland1m features from osm_baseland1m_polygon table.");
            Collection<Feature> baseland1mFeatures = baseland1mFeatureSource.GetAllFeatures(ReturningColumnsType.NoColumns);

            Collection<Feature> newBaselandFeatures = new Collection<Feature>();

            foreach (Feature feature in baselandFeatures)
            {
                UpdateStatus($"Getting intersection feature for baseland feature: {feature.Id}.");

                Feature validFeature = SqlTypesGeometryHelper.MakeValid(feature);
                newBaselandFeatures.Add(validFeature.GetIntersection(boundaryFeature));
            }

            foreach (Feature feature in baseland1mFeatures)
            {
                UpdateStatus($"Getting difference feature for baseland1m feature: {feature.Id}.");
                if (feature.IsDisjointed(boundaryFeature)) newBaselandFeatures.Add(feature);
                else
                {
                    Feature validFeature = SqlTypesGeometryHelper.MakeValid(feature);
                    Feature difference = SqlTypesGeometryHelper.GetDifference(validFeature, boundaryFeature);

                    if (difference != null) newBaselandFeatures.Add(difference);
                }
            }

            baselandFeatureSource.BeginTransaction();
            UpdateStatus("Deleting table osm_baseland_polygon.");
            baselandFeatureSource.ExecuteNonQuery("DELETE FROM osm_baseland_polygon;");
            UpdateStatus("Deleting table idx_osm_baseland_polygon_geometry.");
            baselandFeatureSource.ExecuteNonQuery("DELETE FROM idx_osm_baseland_polygon_geometry;");

            int counter = 0;
            foreach (Feature f in newBaselandFeatures)
            {
                UpdateStatus($"Adding new baseland feature: {f.Id}.");
                Feature feature = new Feature(f.GetWellKnownBinary());
                baselandFeatureSource.AddFeature(feature);
                counter++;
            }
            UpdateStatus("Commiting sqlite database transaction.");
            baselandFeatureSource.CommitTransaction();
        }

        private Collection<View> GetViews(SQLiteConnection connection)
        {
            // Here we query all of the views
            SQLiteCommand getTablesCommand = new SQLiteCommand(@"SELECT name, sql FROM sqlite_master WHERE (type='view');", connection);
            SQLiteDataReader getTablesreader = getTablesCommand.ExecuteReader();

            Collection<View> views = new Collection<View>();

            // Add the views into a collection
            while (getTablesreader.Read())
            {
                string viewName = getTablesreader.GetValue(0).ToString();
                string createStatement = getTablesreader.GetValue(1).ToString();

                //Extract the base table from the create statement
                //Get rid of new lines
                string baseTable = createStatement.Replace("\n", "");
                //Get rid of tabs
                baseTable = baseTable.Replace("\t", "");
                //Get only the text after the "FROM" Statement
                baseTable = baseTable.Substring(baseTable.LastIndexOf("FROM ") + "FROM ".Length);
                //Get only the text before tho "WHERE" statement
                //uses inline if to handle the case of no "WHERE" statement
                //see osm_address_point for an example of this
                baseTable = baseTable.Substring(0, baseTable.LastIndexOf("WHERE") > 0 ? baseTable.LastIndexOf("WHERE") : baseTable.Length);
                //Get rid of any leading or trailing spaces that might exist
                baseTable = baseTable.Trim();
                View view = new View(viewName, baseTable, createStatement);

                views.Add(view);
            }
            return views;
        }

        private Collection<Table> GetTables(SQLiteConnection connection)
        {
            // Here we query all of the tables which are not master tables or the indexes
            SQLiteCommand getTablesCommand = new SQLiteCommand(@"SELECT sqlite_master.name FROM sqlite_master LEFT OUTER JOIN sqlite_temp_master ON sqlite_master.name = sqlite_temp_master.name WHERE (sqlite_master.type = 'table' OR sqlite_temp_master.type = 'table') AND sqlite_master.name NOT LIKE 'sqlite_%' and sqlite_master.name NOT LIKE 'idx_%' and sqlite_master.name NOT LIKE 'planet_%'", connection);
            SQLiteDataReader getTablesreader = getTablesCommand.ExecuteReader();

            Collection<Table> tables = new Collection<Table>();

            // Add the tables into a collection of table objects
            while (getTablesreader.Read())
            {
                Table table = new Table();
                table.TableName = getTablesreader.GetValue(0).ToString();
                table.Columns = new Collection<SqliteColumn>();
                tables.Add(table);
            }

            // Cleanup the table reader and command
            getTablesreader.Close();
            getTablesreader.Dispose();

            // For each table we now get and load the column names and data types
            foreach (Table table in tables)
            {
                // Get the column info for the table
                SQLiteCommand getTableSchemasCommand = new SQLiteCommand(string.Format("Pragma table_info({0});", table.TableName), connection);
                SQLiteDataReader getSchemaReader = getTableSchemasCommand.ExecuteReader();

                while (getSchemaReader.Read())
                {
                    string columnName = getSchemaReader.GetValue(1).ToString();
                    string columnType = getSchemaReader.GetValue(2).ToString();

                    if (!(columnName.Equals("geometry") || columnName.Equals("id")))
                    {
                        switch(columnType)
                        {
                            case "character varying":
                                table.Columns.Add(new SqliteColumn() { ColumnName = columnName, ColumnType = SqliteColumnType.Text });
                                break;
                            case "bigint":
                            case "INT":
                                table.Columns.Add(new SqliteColumn() { ColumnName = columnName, ColumnType = SqliteColumnType.Integer });
                                break;
                            case "double precision":
                                table.Columns.Add(new SqliteColumn() { ColumnName = columnName, ColumnType = SqliteColumnType.Real });
                                break;
                            default:
                                table.Columns.Add(new SqliteColumn() { ColumnName = columnName, ColumnType = (SqliteColumnType)Enum.Parse(typeof(SqliteColumnType), columnType, true) });
                                break;
                        }                                                    
                    }
                }

                // Cleanup the column command and reader
                getSchemaReader.Close();
                getSchemaReader.Dispose();
            }

            return tables;
        }

        private void ExecuteNonQueryCommand(StringBuilder commandText, SQLiteConnection connection)
        {
            ExecuteNonQueryCommand(commandText.ToString(), connection);
        }

        private void ExecuteNonQueryCommand(string commandText, SQLiteConnection connection)
        {
            SQLiteCommand command = new SQLiteCommand(commandText, connection);
            command.ExecuteNonQuery();
            command.Dispose();
        }

        private Collection<Feature> GetFeaturesIntersects(FeatureSource source, BaseShape targetShape, ReturningColumnsType returningColumnsType)
        {
            RectangleShape boundingBox = targetShape.GetBoundingBox();
            Collection<Feature> allPossibleFeatures = source.GetFeaturesInsideBoundingBox(boundingBox, returningColumnsType);

            Collection<Feature> returnFeatures = new Collection<Feature>();
            foreach (Feature feature in allPossibleFeatures)
            {
                ShapeValidationResult validationResult = feature.GetShape().Validate(ShapeValidationMode.Simple);
                if (validationResult.IsValid && feature.GetShape().Intersects(targetShape))
                {
                    returnFeatures.Add(feature);
                }
            }
            return returnFeatures;
        }
    }
}
